
import pandas as pd

from sorted_nearest import annotate_clusters, cluster_by


def _cluster(df, kwargs):

    if df.empty:
        return None


    slack = kwargs.get("slack", 0)

    cdf = df.sort_values("Start")

    ids = annotate_clusters(cdf.Start.values, cdf.End.values, slack)

    cdf.insert(df.shape[1], "Cluster", ids)
    return cdf

def _cluster_by(df, kwargs):

    if df.empty:
        return None

    slack = kwargs.get("slack", 0)

    by = kwargs["by"]
    cdf = df.sort_values(by)

    new_ids = (cdf[by] != cdf[by].shift()).cumsum()
    cdf.insert(cdf.shape[1], "ClusterBy", new_ids)

    cdf = cdf.sort_values(["ClusterBy", "Start"])

    ids = cluster_by(cdf.Start.values, cdf.End.values, cdf.ClusterBy.values, slack)

    cdf = cdf.drop("ClusterBy", axis=1)
    cdf.insert(cdf.shape[1], "Cluster", ids)

    return cdf



import numpy as np
import pandas as pd
from ncls import NCLS


def _number_overlapping(scdf, ocdf, kwargs):

    keep_nonoverlapping = kwargs.get("keep_nonoverlapping", True)

    if scdf.empty:
        return None
    if ocdf.empty:
        if keep_nonoverlapping:
            df = scdf.copy()
            # print(df)
            df.insert(df.shape[1], "NumberOverlaps", 0)
            # print("df" * 100)
            # print(df)
            return df
        else:
            return None

    oncls = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    _self_indexes, _other_indexes = oncls.all_overlaps_both(
        starts, ends, indexes)

    s = pd.Series(_self_indexes)
    counts_per_read = s.value_counts()[s.unique()].reset_index()
    counts_per_read.columns = ["Index", "Count"]

    df = scdf.copy()

    if keep_nonoverlapping:
        _missing_indexes = np.setdiff1d(scdf.index, _self_indexes)
        missing = pd.DataFrame(data={"Index": _missing_indexes, "Count": 0}, index=_missing_indexes)
        counts_per_read = pd.concat([counts_per_read, missing])
    else:
        df = df.loc[_self_indexes]

    counts_per_read = counts_per_read.set_index("Index")

    df.insert(df.shape[1], "NumberOverlaps", counts_per_read)

    return df



def _coverage(scdf, ocdf, kwargs):

    if scdf.empty:
        return None
    if ocdf.empty:
        df = scdf.copy()
        df.insert(df.shape[1], "FractionOverlaps", 0.0)
        return df

    oncls = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    _lengths = oncls.coverage(starts, ends, indexes)
    _lengths = _lengths /  (ends - starts)
    _fractions = _lengths
    _fractions = _fractions.astype("float64")
    _fractions = np.nan_to_num(_fractions)

    scdf = scdf.copy()

    scdf.insert(scdf.shape[1], "FractionOverlaps", _fractions)

    return scdf



import numpy as np
import pandas as pd

from .join import _both_indexes

try:
    dummy = profile
except:
    profile = lambda x: x

np.random.seed(0)


def sort_one_by_one(d, col1, col2):
    """
    Equivalent to pd.sort_values(by=[col1, col2]), but faster.
    """

    d = d.sort_values(by=[col2])
    return d.sort_values(by=[col1], kind='mergesort')


def _insert_distance(ocdf, dist, suffix):

    if "Distance" not in ocdf:
        distance_column_name = "Distance"
    elif "Distance" + suffix not in ocdf:
        distance_column_name = "Distance" + suffix
    else:
        i = 1
        while "Distance" + str(i) in ocdf:
            i += 1
        distance_column_name = "Distance" + str(i)

    ocdf.insert(ocdf.shape[1], distance_column_name,
                pd.Series(dist, index=ocdf.index).fillna(-1).astype(int))

    return ocdf


def nearest_next_idx(d1, d2, k=1):
    """Return k indexes from d1, d2 and their distance.

    d1.End <= d2.Start, i.e. next in forward direction.

    dist negative means invalid, i.e. there were less than k nearest intervals."""

    d1e = d1.End.sort_values()
    d2s = d2.Start.sort_values()

    ix = np.searchsorted(d2s, d1e, side="left")

    if k != 1:
        new_ix = np.ones(len(ix) * (k), dtype=d1e.dtype) * -1
        new_ix[::k] = ix
        for _k in range(1, k):
            _k_m1 = _k - 1
            new_ix[_k::k] = new_ix[_k_m1::k] + 1

        ix = new_ix

    ix[ix >= len(d2s)] = len(d2s) - 1

    d1_idx = d1e.index
    if k != 1:
        r = np.repeat(np.arange(len(d1e)), k)
        d1e = d1e.iloc[r]
        d1_idx = d1e.index

    d2_idx = d2s.iloc[ix].index

    d2s = d2s[d2_idx].values
    dist = d2s - d1e

    return pd.DataFrame({"D1X": d1_idx, "D2X": d2_idx, "Dist": dist})


def nearest_next(d1, d2, kwargs):

    k = kwargs["k"]
    suffix = kwargs["suffix"]

    x = nearest_next_idx(d1, d2, k)
    x = x[x.Dist >= 0]
    d1 = d1.loc[x.D1X]
    d1.index = range(len(d1))
    d2 = d2.loc[x.D2X]
    d2.insert(d2.shape[1], "Distance", x.Dist)
    d2.index = range(len(d2))
    return d1.join(d2, rsuffix=suffix)


def nearest_previous_idx(d1, d2, k=1):

    d1s = d1.Start.sort_values()
    d2e = d2.End.sort_values()

    ix = np.searchsorted(d2e, d1s, side="right") - 1
    ix[ix < 0] = 0

    d2_idx = d2e.iloc[ix].index

    if k != 1:
        new_ix = np.ones(len(ix) * (k), dtype=d1s.dtype) * -1
        new_ix[::k] = ix
        for _k in range(1, k):
            _k_m1 = _k - 1
            new_ix[_k::k] = new_ix[_k_m1::k] - 1

        ix = new_ix

    ix[ix < 0] = 0

    d1_idx = d1s.index
    if k != 1:
        r = np.repeat(np.arange(len(d1s)), k)
        d1s = d1s.iloc[r]
        d1_idx = d1s.index

    d2_idx = d2e.iloc[ix].index

    dist = d1s - d2e[d2_idx].values

    return pd.DataFrame({"D1X": d1_idx, "D2X": d2_idx, "Dist": dist})


def nearest_previous(d1, d2, kwargs):

    suffix = kwargs["suffix"]

    x = nearest_previous_idx(d1, d2, k)
    x = x[x.Dist >= 0]
    d1 = d1.loc[x.D1X]
    d1.index = range(len(d1))
    d2 = d2.loc[x.D2X]
    d2.insert(d2.shape[1], "Distance", x.Dist)
    d2.index = range(len(d2))
    return d1.join(d2, rsuffix=suffix)


def nearest_idx(d1, d2, k=1):

    n = nearest_next_idx(d1, d2, k)
    p = nearest_previous_idx(d1, d2, k)

    df = pd.concat([n, p])
    df = df[df.Dist >= 0]
    df = sort_one_by_one(df, "D1X", "Dist")
    # df = df.groupby("D1X", sort=False).head(k)

    return df


nearest_method = {
    ("upstream", "+"): nearest_previous_idx,
    ("upstream", "-"): nearest_next_idx,
    ("downstream", "+"): nearest_next_idx,
    ("downstream", "-"): nearest_previous_idx
}


@profile
def overlap_for_nearest(scdf, ocdf):

    six, oix = _both_indexes(scdf, ocdf)

    return pd.DataFrame(data={"D1X": six, "D2X": oix, "Dist": 0}, index=six)


@profile
def _nearest(d1, d2, kwargs):

    if d1.empty or d2.empty:
        return None

    how = kwargs["how"]
    k = kwargs["k"]
    suffix = kwargs["suffix"]

    if how == "upstream" or how == "downstream":
        strand = d1.Strand.iloc[0]
        __nearest = nearest_method[how, strand]
    else:
        __nearest = nearest_idx

    x = __nearest(d1, d2)

    x = x[x.Dist >= 0]  # negative dist denotes "had no nearest"
    x.Dist += 1

    x = x.drop_duplicates()

    print("x" * 10)
    print(x)
    if overlap:
        overlapping_ix = overlap_for_nearest(d1, d2)
        print("overlap " * 10)
        print(overlapping_ix)
        x = pd.concat([overlapping_ix, x])

    d1 = d1.loc[x.D1X]
    d1.index = range(len(d1))
    d2 = d2.loc[x.D2X]
    d2.index = range(len(d2))

    j = d1.join(d2, rsuffix=suffix)

    j = _insert_distance(j, x.Dist.values, suffix)

    return j


# def _k_nearest(d1, d2, kwargs):

# def _nearest(d1, d2, kwargs):

#     how = kwargs["how"]
#     k = kwargs["k"]
#     suffix = kwargs["suffix"]
#     overlap = kwargs["overlap"]

#     overlapping, nonoverlapping = d1.overlap_and_nonoverlap_tuple(d2)



import pyranges as pr
import numpy as np

np.random.seed(0)


def test_stranded():
    cpg = pr.data.cpg()
    exons = pr.data.exons()

    j = cpg.join(exons)

    assert j.stranded

    j.Strand = "."

    assert not j.stranded

    j.Strand = np.random.choice("+ -".split(), size=len(j))

    assert j.stranded

    for _, df in j:
        assert len(df.Strand.drop_duplicates()) == 1


def test_unstrand():

    exons = pr.data.exons()

    cpg = pr.data.cpg()
    x = exons.unstrand()
    for _, df in x:
        print(len(df.index), len(set(df.index)))
        assert len(df.index) == len(set(df.index))

    # x = pr.PyRanges(x.df.reset_index(drop=True))
    cpg.join(x)



import shutil

import pyranges as pr
import pandas as pd

from tabulate import tabulate


def get_terminal_size():

    return shutil.get_terminal_size().columns

def show_pos_merge_position(self, df, first_df):

    if "Strand" in df:
        pos = df.Chromosome.astype(str) + " " + df.Start.astype(
            str) + "-" + df.End.astype(str) + " " + df.Strand.astype(str)
    else:
        pos = df.Chromosome.astype(str) + " " + df.Start.astype(
            str) + "-" + df.End.astype(str)

    return pos



def increase_string_width(self, df, first_df, merge_position):

    # TODO: increase until no longer fits the screen
    # instead of decreasing. Takes long time with 60k+ cols

    stranded = self.stranded
    n_intervals = len(self)
    n_chromosomes = len(self.chromosomes)

    terminal_width = get_terminal_size()

    columns = []
    all_dtypes = first_df.dtypes
    if not merge_position:
        if "Strand" in df:
            # if the df is stranded, always include enough cols to show strand
            for c in df.columns:
                columns.append(c)
                if c == "Strand":
                    break
            build_df = df.get(columns)
            never_add = columns[:]
        else:
            columns = "Chromosome Start End".split()
            build_df = df.get(columns)
            never_add = "Chromosome Start End Strand".split()

        dtypes = []
        for c in columns:
            dtype = all_dtypes[c]
            dtypes.append(dtype)
    else:
        build_df = show_pos_merge_position(self, df, first_df)
        columns = ["-Position-"]
        dtypes = ["Multiple types"]
        never_add = "Chromosome Start End Strand".split()

    h = [c + "\n(" + str(t) + ")" for c, t in zip(columns, list(dtypes))]

    str_repr = tabulate(df.get(columns), headers=h, tablefmt='psql', showindex=False)
    table_width = len(str_repr.split("\n", 1)[0])

    for i, c in enumerate(df.columns):
        if c in never_add:
            continue

        columns.append(c)
        t = all_dtypes[c]
        dtypes.append(t)
        _h = c + "\n(" + str(t) + ")"
        h.append(_h)

        new_build_df =  pd.concat([build_df, df[c]], axis=1)

        new_str_repr = tabulate(new_build_df, headers=h, tablefmt='psql', showindex=False)

        table_width = len(new_str_repr.split("\n", 1)[0])
        # print("1", table_width, get_terminal_size())
        if table_width >= terminal_width:
            first_hidden = columns[-1]
            columns = columns[:-1]
            dtypes = dtypes[:-1]
            break
        else:
            str_repr = new_str_repr
            build_df = new_build_df
            # print("2", table_width, get_terminal_size())

    hidden_cols = set(df.columns) - (set(columns).union(never_add))
    n_hidden_cols = len(hidden_cols)
    stranded = "Stranded" if self.stranded else "Unstranded"
    str1 = "{} PyRanges object has {:,} rows and {:,} columns from {} chromosomes.".format(
        stranded, n_intervals, len(self.columns), n_chromosomes)

    if n_hidden_cols:

        # try to add ... as last col
        ddd = pd.Series("...", index=build_df.index)
        ddd.name = "...\n..."
        h[-1] = ddd.name

        new_build_df =  pd.concat([build_df, ddd], axis=1)
        new_str_repr = tabulate(new_build_df, headers=h, tablefmt='psql', showindex=False)
        table_width = len(new_str_repr.split("\n", 1)[0])

        if table_width <= terminal_width:
            str_repr = new_str_repr
        else:
            # need to remove last columns, because adding ... made it too wide
            columns = list(build_df.columns)
            first_hidden = columns[-1]
            build_df = build_df.drop(columns[-1], axis=1)
            ddd = pd.Series("...", index=build_df.index)
            ddd.name = "..."
            h[-2] = ddd.name
            new_build_df =  pd.concat([build_df, ddd], axis=1)
            str_repr = tabulate(new_build_df, headers=h, tablefmt='psql', showindex=False)

        # add hidden col info
        columns = list(df.columns)
        first_hidden_idx = columns.index(first_hidden)
        str2 = "Hidden columns: {}".format(", ".join(columns[first_hidden_idx:first_hidden_idx+10]))
        if (n_hidden_cols - 10) > 0:
            str2 += "... (+ {} more.)".format(n_hidden_cols - 10)
            str_repr = "\n".join([str_repr, str1, str2])
        else:
            str_repr = "\n".join([str_repr, str1, str2])
    else:
        str_repr = "\n".join([str_repr, str1])

    if "Strand" in self.columns and not self.stranded:
        strands = []
        for _, df in self:
            strands.extend(list(df.Strand.drop_duplicates()))

        untraditional_strands = set(strands) - set("+-")
        more_than_10 = ", ..." if len(untraditional_strands) > 9 else ""
        from itertools import islice
        untraditional_strands = islice(untraditional_strands, 10)
        str_repr += "\n(Considered unstranded due to the values {}{} being present in the Strand column.)".format(
            ", ".join((str(x) for x in untraditional_strands)), more_than_10)

    return str_repr


def tostring(self, n=8, merge_position=False):

    entries = n
    half_entries = int(entries / 2)

    # TODO: test this f
    if len(self) == 0:
        return "Empty PyRanges"

    # keys = natsorted(list(self.dfs.keys()))

    if len(self.keys()) == 1:


        first_key = self.keys()[0]
        # last_key = list(self.dfs.keys())[-1]
        first_df = self.dfs[first_key]

        # last_df = ray.get(self.dfs[last_key]).tail(half_entries)
        h = first_df.head(half_entries).astype(object)
        m = first_df.head(1).astype(object)
        t = first_df.tail(half_entries).astype(object)
        m.loc[:, :] = "..."

        if len(self) > entries:
            s = pd.concat([h, m, t])
        elif len(self) == entries:
            s = pd.concat([h, t])
        else: # < entries
            s = first_df.copy()
    elif len(self) <= entries:
        s = self.df.copy()
        s = s.astype(object)
        first_df = s.copy()
    else:
        keys = self.keys()
        first_key = keys[0]
        last_key = keys[-1]
        # first_key = self.keys[0]
        # last_key = self.keys[-1]
        first_df = self.dfs[first_key].head(half_entries)
        last_df = self.dfs[last_key].tail(half_entries)

        # last_df = self.dfs[list(self.dfs.keys())[-1]].tail(half_entries)

        h = first_df.head(half_entries).astype(object)
        m = first_df.head(1).astype(object)
        t = last_df.head(half_entries).astype(object)
        m.loc[:, :] = "..."
        # m.index = ["..."]
        # print((len(h) + len(t)) < entries, len(self) >= entries)
        if (len(h) + len(t)) < entries:

            keys_covered = set()
            # iterate from front until have three
            heads = []
            hl = 0
            for k in keys:
                keys_covered.add(k)
                h = self.dfs[k].head(half_entries)
                first_df = h
                hl += len(h)
                heads.append(h)
                if hl >= half_entries:
                    break

            tails = []
            tl = 0

            for k in keys[::-1]:
                if k in keys_covered:
                    continue

                t = self.dfs[k].tail(half_entries)
                tl += len(t)
                tails.append(t)
                if tl >= half_entries:
                    break
            # iterate from back until have three

            h = pd.concat(heads).head(half_entries).astype(object)
            if tails:
                t = pd.concat(tails).tail(half_entries).astype(object)
                if len(h) + len(t) >= entries:
                    m = h.head(1).astype(object)
                    m.loc[:, :] = "..."
                    s = pd.concat([h, m, t])
                else:
                    s = pd.concat([h, t])
            else:
                s = h

        elif len(h) + len(t) == entries:
            m.loc[:, :] = "..."
            s = pd.concat([h, m, t])
        else:
            s = pd.concat([h, t])

    str_repr = increase_string_width(self, s, first_df, merge_position)

    return str_repr


def sort_tostring(self, n=30, merge_position=False):

    sort_cols = "Start End".split()
    if self.stranded:
        sort_cols.append("Strand")

    first_df = self.values()[0]

    if len(self) <= n:
        df = self.df.sort_values(sort_cols)
    else:
        dfs = []
        half_n = int(n/2)
        current_len = 0

        for chromosome in self.chromosomes:

            df = self[chromosome].df.sort_values(sort_cols)
            current_len += len(df)
            dfs.append(df.head(half_n))

            if current_len > half_n:
                break

        head = pd.concat(dfs)

        current_len = 0
        dfs = []
        for chromosome in reversed(self.chromosomes):

            df = self[chromosome].df.sort_values(sort_cols)
            current_len += len(df)
            dfs.append(df.tail(half_n))

            if current_len > half_n:
                break

        tail = pd.concat(dfs)

        middle = tail.head(1).copy().astype(object)
        middle.loc[:, :] = "..."

        df = pd.concat([head, middle, tail])

    str_repr = increase_string_width(self, df, first_df, merge_position)

    return str_repr




import pyranges as pr
import pandas as pd

from collections import defaultdict

def itergrs(prs, strand=None):

    if strand is None:
        strand =  all([gr.stranded for gr in prs])

    if strand is False and any([gr.stranded for gr in prs]):
        prs = [gr.unstrand() for gr in prs]

    grs_per_chromosome = defaultdict(list)
    set_keys = set()
    for gr in prs:
        set_keys.update(gr.dfs.keys())

    empty_dfs = [pd.DataFrame(columns=gr.columns) for gr in prs]
    for gr, empty in zip(prs, empty_dfs):
        for k in set_keys:
            df = gr.dfs.get(k, empty)
            grs_per_chromosome[k].append(df)

    return iter(grs_per_chromosome.values())



import numpy as np
import pandas as pd

from .join import _both_indexes

from sorted_nearest import nearest_previous_nonoverlapping_all, nearest_next_nonoverlapping_all

try:
    dummy = profile
except:
    profile = lambda x: x

np.random.seed(0)


def nearest_previous_idx(d1, d2, k):

    d1 = d1.sort_values("Start")
    d2 = d2.sort_values("End")

    d1s = d1.Start
    d2e = d2.End

    # ix where d1s could be inserted into d2e to preserve order of d2e
    ix = np.searchsorted(d2e, d1s, side="right") - 1

    valid = ix >= 0
    ix = ix[valid]

    d1s = d1s.iloc[valid]

    lidx, ridx_pos, dist = nearest_previous_nonoverlapping_all(
        d1s.values, d2e.values, d1s.index.values, ix, k)

    ridx = d2e.iloc[ridx_pos].index

    return lidx, ridx, dist


def nearest_next_idx(d1, d2, k):

    d1 = d1.sort_values("End")
    d2 = d2.sort_values("Start")

    d1e = d1.End
    d2s = d2.Start

    # ix where d1e could be inserted into d2s to preserve order of d2s
    ix = np.searchsorted(d2s, d1e, side="left")

    print(ix)
    valid = ix < len(d2s)
    ix = ix[valid]

    d1e = d1e.iloc[valid]

    print(ix)
    print(d1)
    print(d1e)
    print(d2)

    lidx, ridx_pos, dist = nearest_next_nonoverlapping_all(
        d1e.values, d2s.values, d1e.index.values, ix, k)

    ridx = d2s.iloc[ridx_pos].index

    return lidx, ridx, dist


def nearest_previous(d1, d2, kwargs):

    k = kwargs.get("k", 1)
    suffix = kwargs.get("suffix", "_b")

    lidx, ridx, dist = nearest_previous_idx(d1, d2, k)

    d1 = d1.reindex(lidx)
    d2 = d2.reindex(ridx)
    d1.index = range(len(d1))
    d2.index = range(len(d1))
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", dist)
    return df


def nearest_next(d1, d2, kwargs):

    k = kwargs.get("k", 1)
    suffix = kwargs.get("suffix", "_b")

    lidx, ridx, dist = nearest_next_idx(d1, d2, k)

    d1 = d1.reindex(lidx)
    d2 = d2.reindex(ridx)
    d1.index = range(len(d1))
    d2.index = range(len(d1))
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", dist)
    return df


@profile
def _nearest(d1, d2, kwargs):

    if d1.empty or d2.empty:
        return None

    how = kwargs["how"]
    suffix = kwargs["suffix"]
    k = kwargs["k"]

    # return nearest_previous(d1, d2, kwargs)
    return nearest_next(d1, d2, kwargs)
    # lidx, ridx, dist = nearest_previous_idx(d1, d2, k)

    # if how == "upstream" or how == "downstream":
    #     strand = d1.Strand.iloc[0]
    #     __nearest = nearest_method[how, strand]
    # else:
    #     __nearest = nearest_idx

    # x = __nearest(d1, d2)

    # x = x[x.Dist >= 0]  # negative dist denotes "had no nearest"
    # x.Dist += 1

    # x = x.drop_duplicates()

    # print("x" * 10)
    # print(x)
    # if overlap:
    #     overlapping_ix = overlap_for_nearest(d1, d2)
    #     print("overlap " * 10)
    #     print(overlapping_ix)
    #     x = pd.concat([overlapping_ix, x])

    # d1 = d1.loc[x.D1X]
    # d1.index = range(len(d1))
    # d2 = d2.loc[x.D2X]
    # d2.index = range(len(d2))

    # j = d1.join(d2, rsuffix=suffix)

    # j = _insert_distance(j, x.Dist.values, suffix)

    # return j



from io import StringIO
import pandas as pd
import pyranges as pr
import pytest

from hypothesis import given, settings, HealthCheck  #, assume
from hypothesis import reproduce_failure  # pylint: disable=unused-import

from tests.hypothesis_helper import genomicfeature

from tests.hypothesis_helper import max_examples, deadline, slow_max_examples

by_to_id = {"gene": "gene_id", "transcript": "transcript_id"}

def compute_introns_single(df, by):
    id_column = by_to_id[by]
    df = df.sort_values("Start End".split())
    g = df[df.Feature == by]
    x = pr.PyRanges(df[df.Feature == "exon"]).merge(by=id_column, strand=False)
    if not len(x) or not len(g):
        return
    x.Strand = "-"
    x = x.df

    print("g " * 100)
    print(g)
    print("x " * 100)
    print(x)

    if g.empty or x.empty:
        return pd.DataFrame()

    x = pd.concat([x, x.tail(1)], sort=False).reset_index(drop=True)
    cols = list(x.columns)
    original_cols = cols[::1]
    cols[2] = "Start"
    cols[1] = "End"
    x = x[cols]
    x.columns = original_cols
    x.loc[:, "Start"] = x.Start.shift()
    x.at[x.index[0], "Start"] = g.Start.iloc[0]
    x.at[x.index[-1], "End"] = g.End.iloc[0]
    x.loc[:, "Start"] = x.Start.astype(int)
    # if x.Start.iloc[0] == g.Start.iloc[0]:
    #     print("Start equals")
    # x = x.drop(x.head(1).index)
    # if x.End.iloc[-1] == g.End.iloc[0]:
    #     print("End equals")
    #     x = x.drop(x.head(1).index)
    # x = x[~(x.Start == g.Start.iloc[0])]
    # x = x[~(x.End == g.End.iloc[0])]
    x = x[x.Start != x.End]

    return x.sort_values("Start End".split())#.reset_index(drop=True)

# actual   Chromosome     Start       End Strand
# 0       chr1  13163233  13164184      -
# 1       chr1  13164760  13165156      -
# 2       chr1  13165468  13167115      -
# 3       chr1  13167193  13194517      -
# 4       chr1  13194595  13196244      -
# 5       chr1  13196556  13196952      -
# 6       chr1  13197528  13198478      -
# expected   Chromosome     Start       End Strand
# 5       chr1  13163233  13164184      -
# 6       chr1  13164760  13165156      -
# 7       chr1  13165468  13167115      -
# 8       chr1  13167193  13199727      -
# 2       chr1  13196556  13196952      +
# 3       chr1  13197528  13198478      +
# 4       chr1  13199727  13161985      -



def _introns_correct(full, genes, exons, introns, by):


    """Testing that introns:

    1: ends larger than starts
    2: the intersection of the computed introns and exons per gene are 0
    3: that the number of introns overlapping each gene is the same as number of introns per gene
    4 & 5: that the intron positions are the same as the ones computed with the slow, but correct algo"""
    id_column = by_to_id[by]
    if len(introns):
        assert (introns.Start < introns.End).all(), str(introns[(introns.Start >= introns.End)])

    expected_results = {}
    based_on = {}
    for gene_id, gdf in full.groupby(id_column): # #[full.gene_id.isin(["ENSG00000078808.16"])]
        # print("gdf " * 10)
        # print(gdf)
        if not len(gdf[gdf.Feature == "gene"]) or not len(gdf[gdf.Feature == "transcript"]):
            continue
        expected_results[gene_id] = compute_introns_single(gdf, by)

        based_on[gene_id] = pr.PyRanges(gdf[gdf.Feature.isin([by, "exon"])]).df

    if not len(introns):
        for v in expected_results.values():
            assert v.empty
        return # test passed

    for gene_id, idf in introns.groupby(id_column):
        idf = idf.sort_values("Start End".split())
        if not gene_id in expected_results:
            continue
        expected = expected_results[gene_id]
        exons = pr.PyRanges(based_on[gene_id]).subset(lambda df: df.Feature == "exon").merge(by=id_column)
        genes = pr.PyRanges(based_on[gene_id]).subset(lambda df: df.Feature == by)
        print("exons", exons)
        print("based_on", based_on[gene_id])
        print("actual", idf["Chromosome Start End Strand".split()])
        print("expected", expected["Chromosome Start End Strand".split()])
        _introns = pr.PyRanges(idf)
        assert len(exons.intersect(_introns)) == 0
        assert len(genes.intersect(_introns)) == len(_introns)
        assert list(idf.Start) == list(expected.Start), "not equal"
        assert list(idf.End) == list(expected.End), "not equal"

by = ["gene", "transcript"]

# @pytest.mark.parametrize("by", by)
# @settings(
#     max_examples=slow_max_examples,
#     deadline=deadline,
#     suppress_health_check=HealthCheck.all())
# @given(gr=genomicfeature())  # pylint: disable=no-value-for-parameter
# # @reproduce_failure('4.15.0', b'AAIAAAABAP8A')
# # @reproduce_failure('4.15.0', b'AAIAAAABAOwBAO0A')
# def test_introns(gr, by):

#     id_column = by_to_id[by]

#     delim = "-" * 20
#     print("\n" + delim +" new test " + delim)


#     if not len(gr[gr.Feature == "gene"]) or not len(gr[gr.Feature == "exon"]):
#         pass

#     else:

#         genes = gr[gr.Feature.isin(["gene"])].df
#         exons = gr[gr.Feature.isin(["exon"])].merge(by=id_column).df
#         introns = gr.features.introns(by=by).df

#         print(genes)
#         print(exons)
#         print(introns)

#         _introns_correct(gr.df, genes, exons, introns, by)

def test_introns_single():

    "Assert that our fast method of computing introns is the same as the slow, correct one in compute_introns_single"

    gr = pr.data.gencode_gtf()[["gene_id", "Feature"]]
    exons = gr[gr.Feature == "exon"].merge(by="gene_id")
    exons.Feature = "exon"
    exons = exons.df
    df = pd.concat([gr[gr.Feature == "gene"].df, exons], sort=False)
    print(df)

    for gid, gdf in df.groupby("gene_id"):
        print("-------" * 20)
        print(gid)
        print(gdf)
        print("gdf", len(gdf))
        expected = compute_introns_single(gdf, by="gene")
        print("expected", len(expected))
        actual = pr.PyRanges(gdf).features.introns().df
        print("actual", len(actual))
        if actual.empty:
            assert expected.empty
            continue

        assert list(expected.Start) == list(actual.Start)
        assert list(expected.End) == list(actual.End)



import numpy as np

import pandas as pd

import pyranges as pr

from natsort import natsorted

import sys
import os


def get_n_args(f):

    import inspect
    nparams = len(inspect.signature(f).parameters)
    return nparams


def call_f(f, nparams, df, odf, kwargs):

    if nparams == 3:
        return f.remote(df, odf, kwargs)
    else:
        return f.remote(df, odf)


def call_f_single(f, nparams, df, kwargs):

    if nparams == 2:
        return f.remote(df, kwargs)
    else:
        return f.remote(df)


class suppress_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    '''

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = (os.dup(1), os.dup(2))

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        os.close(self.null_fds[0])
        os.close(self.null_fds[1])


import pandas as pd

import pyranges as pr

from natsort import natsorted

import sys


def merge_dfs(df1, df2):

    if not df1.empty and not df2.empty:
        return pd.concat([df1, df2], sort=False).reset_index(drop=True)

    elif df1.empty and df2.empty:
        # can this happen?
        return None
    elif df1.empty:
        return df2
    else:
        return df1


def process_results(results, keys):

    results_dict = {k: r for k, r in zip(keys, results) if r is not None}

    try:
        first_item = next(iter(results_dict.values()))
    except StopIteration:  # empty collection
        return results_dict

    if not isinstance(first_item, pd.DataFrame):
        return results_dict

    to_delete = []
    # to ensure no duplicate indexes and no empty dataframes
    for k in results_dict:
        if results_dict[k] is None or results_dict[k].empty:
            to_delete.append(k)
        else:
            # pandas might make a df that is not always C-contiguous
            # copying fixes this
            # TODO: better to only fix columns that are not C-contiguous?
            results_dict[k] = results_dict[k].copy(deep=True)
            results_dict[k].index = range(len(results_dict[k]))

    for k in to_delete:
        del results_dict[k]

    return results_dict


def make_sparse(df):

    if "Strand" in df:
        cols = "Chromosome Start End Strand".split()
    else:
        cols = "Chromosome Start End".split()

    return df[cols]


def make_binary_sparse(kwargs, df, odf):

    sparse = kwargs.get("sparse")

    if not sparse:
        return df, odf

    if sparse.get("self"):

        df = make_sparse(df)

    if sparse.get("other"):

        odf = make_sparse(odf)

    return df, odf


def make_unary_sparse(kwargs, df):

    sparse = kwargs.get("sparse").get("self")

    if sparse:
        df = make_sparse(df)

    return df


def ray_initialized():
    def test_function():
        pass

    try:
        test_function = ray.remote(test_function)
    except Exception as e:
        if type(e) == NameError:
            return False

        raise e

    try:
        test_function.remote()
    except Exception as e:
        if "RayConnectionError" in str(type(e)):
            return True
        else:
            raise e


def get_multithreaded_funcs(function, nb_cpu):

    if nb_cpu > 1:
        import ray
        _merge_dfs = ray.remote(merge_dfs)
        get = ray.get
        function = ray.remote(function)
    else:
        _merge_dfs = lambda: "dummy value"
        _merge_dfs.remote = merge_dfs
        get = lambda x: x
        function.remote = function

    return function, get, _merge_dfs


def pyrange_apply(function, self, other, **kwargs):

    nparams = get_n_args(function)
    nb_cpu = kwargs.get("nb_cpu", 1)

    if nb_cpu > 1:
        import ray
        with suppress_stdout_stderr():
            ray.init(num_cpus=nb_cpu)

    function, get, _merge_dfs = get_multithreaded_funcs(
        function, nb_cpu=nb_cpu)

    strandedness = kwargs["strandedness"]

    other_strand = {"+": "-", "-": "+"}
    same_strand = {"+": "+", "-": "-"}

    if strandedness == "opposite":
        strand_dict = other_strand
    else:
        strand_dict = same_strand

    assert strandedness in ["same", "opposite", False, None]

    if strandedness:
        assert self.stranded and other.stranded, \
            "Can only do stranded operations when both PyRanges contain strand info"

    results = []

    items = natsorted(self.dfs.items())
    keys = natsorted(self.dfs.keys())

    if strandedness:

        for (c, s), df in items:

            os = strand_dict[s]

            if not (c, os) in other.keys() or len(other[c, os].values()) == 0:
                odf = pd.DataFrame(columns="Chromosome Start End".split())
            else:
                odf = other[c, os].values()[0]

            df, odf = make_binary_sparse(kwargs, df, odf)
            result = call_f(function, nparams, df, odf, kwargs)

            results.append(result)

    else:

        if self.stranded and not other.stranded:

            for (c, s), df in items:

                if not c in other.chromosomes:
                    odf = pd.DataFrame(columns="Chromosome Start End".split())
                else:
                    odf = other.dfs[c]

                df, odf = make_binary_sparse(kwargs, df, odf)
                result = call_f(function, nparams, df, odf, kwargs)
                results.append(result)

        elif not self.stranded and other.stranded:

            for c, df in items:

                if not c in other.chromosomes:
                    odf = pd.DataFrame(columns="Chromosome Start End".split())
                else:
                    odf1 = other[c, "+"].df
                    odf2 = other[c, "-"].df

                    odf = _merge_dfs.remote(odf1, odf2)

                df, odf = make_binary_sparse(kwargs, df, odf)

                result = call_f(function, nparams, df, odf, kwargs)
                results.append(result)

        elif self.stranded and other.stranded:

            for (c, s), df in self.items():

                if not c in other.chromosomes:
                    odfs = pr.PyRanges(
                        pd.DataFrame(columns="Chromosome Start End".split()))
                else:
                    odfs = other[c].values()

                # from pydbg import dbg
                # dbg(odfs)

                if len(odfs) == 2:
                    odf = _merge_dfs.remote(*odfs)
                elif len(odfs) == 1:
                    odf = odfs[0]
                else:
                    odf = pd.DataFrame(columns="Chromosome Start End".split())

                df, odf = make_binary_sparse(kwargs, df, odf)

                # dbg(df)
                # dbg(odf)

                result = call_f(function, nparams, df, odf, kwargs)
                results.append(result)

        else:

            for c, df in items:
                if not c in other.chromosomes:
                    odf = pd.DataFrame(columns="Chromosome Start End".split())
                else:
                    odf = other.dfs[c]

                df, odf = make_binary_sparse(kwargs, df, odf)

                result = call_f(function, nparams, df, odf, kwargs)
                results.append(result)

    results = get(results)

    results = process_results(results, keys)

    if nb_cpu > 1:
        ray.shutdown()

    return results


def pyrange_apply_single(function, self, strand, kwargs):

    nparams = get_n_args(function)
    nb_cpu = kwargs.get("nb_cpu", 1)

    if nb_cpu > 1:
        import ray
        with suppress_stdout_stderr():
            ray.init(num_cpus=nb_cpu)

    function, get, _merge_dfs = get_multithreaded_funcs(
        function, nb_cpu=nb_cpu)

    if strand:
        assert self.stranded, \
            "Can only do stranded operation when PyRange contains strand info"

    results = []

    if strand:

        for (c, s), df in self.items():

            kwargs["chromosome"] = c
            _strand = s
            kwargs["strand"] = _strand

            df = make_unary_sparse(kwargs, df)
            result = call_f_single(function, nparams, df, kwargs)
            results.append(result)

        keys = self.keys()

    elif not self.stranded:

        keys = []
        for c, df in self.items():

            kwargs["chromosome"] = c

            df = make_unary_sparse(kwargs, df)
            result = call_f_single(function, nparams, df, kwargs)
            results.append(result)
            keys.append(c)

    else:

        keys = []
        for c in self.chromosomes:

            kwargs["chromosome"] = c

            dfs = self[c]

            if len(dfs.keys()) == 2:
                df, df2 = dfs.values()
                # merge strands
                df = _merge_dfs.remote(df, df2)
            else:
                df = dfs.values()[0]

            df = make_unary_sparse(kwargs, df)
            result = call_f_single(function, nparams, df, kwargs)
            results.append(result)
            keys.append(c)

    results = get(results)

    if nb_cpu > 1:
        ray.shutdown()

    results = process_results(results, keys)

    return results


def _lengths(df):

    lengths = df.End - df.Start

    return lengths


def _tss(df, kwargs):

    slack = kwargs["slack"]

    tss_pos = df.loc[df.Strand == "+"]

    tss_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tss_neg.loc[:, "Start"] = tss_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tss = pd.concat([tss_pos, tss_neg], sort=False)
    tss["End"] = tss.Start
    tss.End = tss.End + 1 + slack
    tss.Start = tss.Start - slack
    tss.loc[tss.Start < 0, "Start"] = 0

    return tss.reindex(df.index)


def _tes(df, kwargs):

    slack = kwargs["slack"]

    tes_pos = df.loc[df.Strand == "+"]

    tes_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tes_neg.loc[:, "Start"] = tes_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tes = pd.concat([tes_pos, tes_neg], sort=False)
    tes["Start"] = tes.End
    tes.End = tes.End + 1 + slack
    tes.Start = tes.Start - slack
    tes.loc[tes.Start < 0, "Start"] = 0

    return tes.reindex(df.index)


def _slack(df, kwargs):

    df = df.copy()
    # dtype = df.Start.dtype
    slack = kwargs["slack"]

    assert type(slack) == int, "Slack parameter must be integer, is {}".format(
        type(slack))
    # df = self.df.copy()
    df.loc[:, "Start"] = df.Start - slack
    df.loc[df.Start < 0, "Start"] = 0
    df.End = df.End + slack
    cs = "Start End".split()
    # df[cs] = df[cs].astype(dtype)

    return df


def pyrange_apply_chunks(function, self, as_pyranges, kwargs):

    nparams = get_n_args(function)
    nb_cpu = kwargs.get("nb_cpu", 1)
    if nb_cpu > 1:
        import ray
        with suppress_stdout_stderr():
            ray.init(num_cpus=nb_cpu)

    function, get, _merge_dfs = get_multithreaded_funcs(
        function, nb_cpu=nb_cpu)

    keys = []
    lengths = []
    results = []
    for k, v in self.items():
        dfs = np.array_split(v, nb_cpu)
        lengths.append(len(dfs))
        results.extend(
            [call_f_single(function, nparams, df, kwargs) for df in dfs])
        keys.append(k)

    results = get(results)

    _results = []
    start = 0
    for key, length in zip(keys, lengths):
        end = start + length
        _r = results[start:end]

        if as_pyranges:
            _results.append(pd.concat(_r))
        else:
            _results.append(_r)

        start = end

    results = _results
    if nb_cpu > 1:
        ray.shutdown()

    results = process_results(results, keys)

    return results



from __future__ import print_function

import sys

import pandas as pd

from pyranges.pyranges import PyRanges


from pyranges.version import __version__


def read_bed(f, output_df=False, nrows=None):

    columns = "Chromosome Start End Name Score Strand ThickStart ThickEnd ItemRGB BlockCount BlockSizes BlockStarts".split(
    )

    if f.endswith(".gz"):
        import gzip
        first_start = gzip.open(f).readline().split()[1]
    else:
        first_start = open(f).readline().split()[1]

    header = None

    try:
        int(first_start)
    except ValueError:
        header = 0

    df = pd.read_csv(
        f,
        dtype={
            "Chromosome": "category",
            "Strand": "category"
        },
        nrows=nrows,
        header=header,
        sep="\t")

    df.columns = columns[:df.shape[1]]

    if not output_df:
        return PyRanges(df)
    else:
        return df


def read_bam(f, output_df=False, mapq=0, required_flag=0, filter_flag=1540):

    try:
        import bamread
    except ModuleNotFoundError:
        print("bamread must be installed to read bam. Use `conda install -c bioconda bamread` or `pip install bamread` to install it.")
        sys.exit(1)


    df = bamread.read_bam(f, mapq, required_flag, filter_flag)

    if output_df:
        return df
    else:
        return PyRanges(df)

    return bamread.read_bam(f, mapq, required_flag, filter_flag)


def _fetch_gene_transcript_exon_id(attribute, annotation):

    no_quotes = attribute.str.replace('"', '').str.replace("'", "")

    df = no_quotes.str.extract(
        "gene_id.?(.+?);(?:.*transcript_id.?(.+?);)?(?:.*exon_number.?(.+?);)?(?:.*exon_id.?(.+?);)?",
        expand=True)  # .iloc[:, [1, 2, 3]]

    df.columns = "GeneID TranscriptID ExonNumber ExonID".split()

    # df.loc[:, "ExonNumber"] = df.ExonNumber.astype(int)

    if annotation == "ensembl":
        newdf = []
        for c in "GeneID TranscriptID ExonID".split():
            r = df[c].astype(str).str.extract('(\d+)').astype(float)
            newdf.append(r)

        newdf = pd.concat(newdf, axis=1)
        newdf.insert(2, "ExonNumber", df["ExonNumber"])
        df = newdf

    return df


def skiprows(f):

    try:
        import gzip
        fh = gzip.open(f)
        for i, l in enumerate(fh):
            if l.decode()[0] != "#":
                break
    except (OSError, TypeError):  # not a gzipped file, or StringIO
        fh = open(f)
        for i, l in enumerate(fh):
            if l[0] != "#":
                break

    fh.close()

    return i


def read_gtf(f, full=True, annotation=None, output_df=False, nrows=None):

    _skiprows = skiprows(f)

    if full:
        return read_gtf_full(f, annotation, output_df, nrows, _skiprows)
    else:
        return read_gtf_restricted(f, annotation, output_df, nrows, _skiprows)


def read_gtf_full(f, annotation=None, output_df=False, nrows=None, skiprows=0):
    """seqname - name of the chromosome or scaffold; chromosome names can be given with or without the 'chr' prefix. Important note: the seqname must be one used within Ensembl, i.e. a standard chromosome name or an Ensembl identifier such as a scaffold ID, without any additional content such as species or assembly. See the example GFF output below.
    # source - name of the program that generated this feature, or the data source (database or project name)
    feature - feature type name, e.g. Gene, Variation, Similarity
    start - Start position of the feature, with sequence numbering starting at 1.
    end - End position of the feature, with sequence numbering starting at 1.
    score - A floating point value.
    strand - defined as + (forward) or - (reverse).
    # frame - One of '0', '1' or '2'. '0' indicates that the first base of the feature is the first base of a codon, '1' that the second base is the first base of a codon, and so on..
    attribute - A semicolon-separated list of tag-value pairs, providing additional information about each feature."""
    dtypes = {
        "Chromosome": "category",
        "Feature": "category",
        "Strand": "category"
    }

    names = "Chromosome Source Feature Start End Score Strand Frame Attribute".split(
    )
    # names = "Chromosome Start End Score Strand Source Feature Frame Attribute".split()
    df_iter = pd.read_csv(
        f,
        sep="\t",
        header=None,
        names=names,
        dtype=dtypes,
        chunksize=int(1e5),
        skiprows=skiprows,
        nrows=nrows)

    dfs = []
    for df in df_iter:
        extra = to_rows(df.Attribute)
        df = df.drop("Attribute", axis=1)
        ndf = pd.concat([df, extra], axis=1, sort=False)
        dfs.append(ndf)

    df = pd.concat(dfs, sort=False)

    if not output_df:
        return PyRanges(df)
    else:
        return df


def to_rows(anno):
    rowdicts = []
    for l in anno:
        l = l.replace('"', '').replace(";", "").split()
        rowdicts.append({k: v for k, v in zip(*([iter(l)] * 2))})

    return pd.DataFrame.from_dict(rowdicts).set_index(anno.index)


def read_gtf_restricted(f,
                        annotation=None,
                        output_df=False,
                        skiprows=0,
                        nrows=None):
    """seqname - name of the chromosome or scaffold; chromosome names can be given with or without the 'chr' prefix. Important note: the seqname must be one used within Ensembl, i.e. a standard chromosome name or an Ensembl identifier such as a scaffold ID, without any additional content such as species or assembly. See the example GFF output below.
    # source - name of the program that generated this feature, or the data source (database or project name)
    feature - feature type name, e.g. Gene, Variation, Similarity
    start - Start position of the feature, with sequence numbering starting at 1.
    end - End position of the feature, with sequence numbering starting at 1.
    score - A floating point value.
    strand - defined as + (forward) or - (reverse).
    # frame - One of '0', '1' or '2'. '0' indicates that the first base of the feature is the first base of a codon, '1' that the second base is the first base of a codon, and so on..
    attribute - A semicolon-separated list of tag-value pairs, providing additional information about each feature."""
    dtypes = {
        "Chromosome": "category",
        "Feature": "category",
        "Strand": "category"
    }

    df_iter = pd.read_csv(
        f,
        sep="\t",
        comment="#",
        usecols=[0, 2, 3, 4, 5, 6, 8],
        header=None,
        names="Chromosome Feature Start End Score Strand Attribute".split(),
        dtype=dtypes,
        chunksize=int(1e5),
        nrows=nrows)

    dfs = []
    for df in df_iter:
        # Since Start is 1-indexed
        df.Start -= 1

        if sum(df.Score == ".") == len(df):
            cols_to_concat = "Chromosome Start End Strand Feature".split()
        else:
            cols_to_concat = "Chromosome Start End Strand Feature Score".split(
            )

        extract = _fetch_gene_transcript_exon_id(df.Attribute, annotation)
        extract.columns = "GeneID TranscriptID ExonNumber ExonID".split()

        extract.ExonNumber = extract.ExonNumber.astype(float)

        df = pd.concat([df[cols_to_concat], extract], axis=1, sort=False)

        dfs.append(df)

    df = pd.concat(dfs, sort=False)

    if not output_df:
        return PyRanges(df)
    else:
        return df
